"""Provides the HTTP request handling interface."""
import aiohttp
import asyncio
from .const import __version__, TIMEOUT
from .exceptions import InvalidInvocation, RequestException


class Requestor(object):
    """Requestor provides an interface to HTTP requests."""

    def __getattr__(self, attribute):  # pragma: no cover
        """Pass all undefined attributes to the _http attribute."""
        if attribute.startswith("__"):
            raise AttributeError
        return getattr(self._http, attribute)

    def __init__(
        self,
        user_agent,
        oauth_url="https://oauth.reddit.com",
        reddit_url="https://www.reddit.com",
        session=None,
        loop=None,
    ):
        """Create an instance of the Requestor class.

        :param user_agent: The user-agent for your application. Please follow reddit's
            user-agent guidelines: https://github.com/reddit/reddit/wiki/API#rules
        :param oauth_url: (Optional) The URL used to make OAuth requests to the reddit
            site. (Default: https://oauth.reddit.com)
        :param reddit_url: (Optional) The URL used when obtaining access tokens.
            (Default: https://www.reddit.com)
        :param session: (Optional) A session to handle requests, compatible with
            aiohttp.ClientSession(). (Default: None)

        """
        if user_agent is None or len(user_agent) < 7:
            raise InvalidInvocation("user_agent is not descriptive")

        self.loop = loop or asyncio.get_event_loop()
        self._http = session or aiohttp.ClientSession(
            loop=self.loop, timeout=aiohttp.ClientTimeout(total=None)
        )
        self._http._default_headers[
            "User-Agent"
        ] = f"{user_agent} asyncprawcore/{__version__}"

        self.oauth_url = oauth_url
        self.reddit_url = reddit_url

    async def close(self):
        """Call close on the underlying session."""
        return await self._http.close()

    async def request(self, *args, timeout=TIMEOUT, **kwargs):
        """Issue the HTTP request capturing any errors that may occur."""
        try:
            return await self._http.request(*args, timeout=timeout, **kwargs)
        except Exception as exc:
            raise RequestException(exc, args, kwargs)



"""Test for asyncprawcore.auth.Authorizer classes."""
import asyncprawcore
import asynctest
from asyncprawcore.requestor import Requestor
from .conftest import (
    CLIENT_ID,
    CLIENT_SECRET,
    PASSWORD,
    PERMANENT_GRANT_CODE,
    REDIRECT_URI,
    REFRESH_TOKEN,
    TEMPORARY_GRANT_CODE,
    USERNAME,
    VCR,
)


class AuthorizerTestBase(asynctest.TestCase):
    async def setUp(self):
        self.requestor = Requestor("asyncprawcore:test (by /u/bboe)")
        self.authentication = asyncprawcore.TrustedAuthenticator(
            self.requestor, CLIENT_ID, CLIENT_SECRET
        )

    async def tearDown(self) -> None:
        await self.requestor.close()


class AuthorizerTest(AuthorizerTestBase):
    async def test_authorize__with_permanent_grant(self):
        self.authentication.redirect_uri = REDIRECT_URI
        authorizer = asyncprawcore.Authorizer(self.authentication)
        with VCR.use_cassette("Authorizer_authorize__with_permanent_grant"):
            await authorizer.authorize(PERMANENT_GRANT_CODE)

        self.assertIsNotNone(authorizer.access_token)
        self.assertIsNotNone(authorizer.refresh_token)
        self.assertIsInstance(authorizer.scopes, set)
        self.assertTrue(len(authorizer.scopes) > 0)
        self.assertTrue(authorizer.is_valid())

    async def test_authorize__with_temporary_grant(self):
        self.authentication.redirect_uri = REDIRECT_URI
        authorizer = asyncprawcore.Authorizer(self.authentication)
        with VCR.use_cassette("Authorizer_authorize__with_temporary_grant"):
            await authorizer.authorize(TEMPORARY_GRANT_CODE)

        self.assertIsNotNone(authorizer.access_token)
        self.assertIsNone(authorizer.refresh_token)
        self.assertIsInstance(authorizer.scopes, set)
        self.assertTrue(len(authorizer.scopes) > 0)
        self.assertTrue(authorizer.is_valid())

    async def test_authorize__with_invalid_code(self):
        self.authentication.redirect_uri = REDIRECT_URI
        authorizer = asyncprawcore.Authorizer(self.authentication)
        with VCR.use_cassette("Authorizer_authorize__with_invalid_code"):
            with self.assertRaises(asyncprawcore.OAuthException):
                await authorizer.authorize("invalid code")
        self.assertFalse(authorizer.is_valid())

    async def test_authorize__fail_without_redirect_uri(self):
        authorizer = asyncprawcore.Authorizer(self.authentication)
        with self.assertRaises(asyncprawcore.InvalidInvocation):
            await authorizer.authorize("dummy code")
        self.assertFalse(authorizer.is_valid())

    def test_initialize(self):
        authorizer = asyncprawcore.Authorizer(self.authentication)
        self.assertIsNone(authorizer.access_token)
        self.assertIsNone(authorizer.scopes)
        self.assertIsNone(authorizer.refresh_token)
        self.assertFalse(authorizer.is_valid())

    def test_initialize__with_refresh_token(self):
        authorizer = asyncprawcore.Authorizer(
            self.authentication, refresh_token=REFRESH_TOKEN
        )
        self.assertIsNone(authorizer.access_token)
        self.assertIsNone(authorizer.scopes)
        self.assertEqual(REFRESH_TOKEN, authorizer.refresh_token)
        self.assertFalse(authorizer.is_valid())

    def test_initialize__with_untrusted_authenticator(self):
        authenticator = asyncprawcore.UntrustedAuthenticator(None, None)
        authorizer = asyncprawcore.Authorizer(authenticator)
        self.assertIsNone(authorizer.access_token)
        self.assertIsNone(authorizer.scopes)
        self.assertIsNone(authorizer.refresh_token)
        self.assertFalse(authorizer.is_valid())

    async def test_refresh(self):
        authorizer = asyncprawcore.Authorizer(
            self.authentication, refresh_token=REFRESH_TOKEN
        )
        with VCR.use_cassette("Authorizer_refresh"):
            await authorizer.refresh()

        self.assertIsNotNone(authorizer.access_token)
        self.assertIsInstance(authorizer.scopes, set)
        self.assertTrue(len(authorizer.scopes) > 0)
        self.assertTrue(authorizer.is_valid())

    async def test_refresh__post_refresh_callback(self):
        def callback(authorizer):
            self.assertNotEqual(REFRESH_TOKEN, authorizer.refresh_token)
            authorizer.refresh_token = "manually_updated"

        authorizer = asyncprawcore.Authorizer(
            self.authentication,
            post_refresh_callback=callback,
            refresh_token=REFRESH_TOKEN,
        )
        with VCR.use_cassette("Authorizer_refresh"):
            await authorizer.refresh()

        self.assertIsNotNone(authorizer.access_token)
        self.assertEqual("manually_updated", authorizer.refresh_token)
        self.assertIsInstance(authorizer.scopes, set)
        self.assertTrue(len(authorizer.scopes) > 0)
        self.assertTrue(authorizer.is_valid())

    async def test_refresh__post_refresh_callback__async(self):
        async def callback(authorizer):
            self.assertNotEqual(REFRESH_TOKEN, authorizer.refresh_token)
            authorizer.refresh_token = "manually_updated"

        authorizer = asyncprawcore.Authorizer(
            self.authentication,
            post_refresh_callback=callback,
            refresh_token=REFRESH_TOKEN,
        )
        with VCR.use_cassette("Authorizer_refresh"):
            await authorizer.refresh()

        self.assertIsNotNone(authorizer.access_token)
        self.assertEqual("manually_updated", authorizer.refresh_token)
        self.assertIsInstance(authorizer.scopes, set)
        self.assertTrue(len(authorizer.scopes) > 0)
        self.assertTrue(authorizer.is_valid())

    async def test_refresh__pre_refresh_callback(self):
        def callback(authorizer):
            self.assertIsNone(authorizer.refresh_token)
            authorizer.refresh_token = REFRESH_TOKEN

        authorizer = asyncprawcore.Authorizer(
            self.authentication, pre_refresh_callback=callback
        )
        with VCR.use_cassette("Authorizer_refresh"):
            await authorizer.refresh()

        self.assertIsNotNone(authorizer.access_token)
        self.assertIsInstance(authorizer.scopes, set)
        self.assertTrue(len(authorizer.scopes) > 0)
        self.assertTrue(authorizer.is_valid())

    async def test_refresh__pre_refresh_callback__async(self):
        async def callback(authorizer):
            self.assertIsNone(authorizer.refresh_token)
            authorizer.refresh_token = REFRESH_TOKEN

        authorizer = asyncprawcore.Authorizer(
            self.authentication, pre_refresh_callback=callback
        )
        with VCR.use_cassette("Authorizer_refresh"):
            await authorizer.refresh()

        self.assertIsNotNone(authorizer.access_token)
        self.assertIsInstance(authorizer.scopes, set)
        self.assertTrue(len(authorizer.scopes) > 0)
        self.assertTrue(authorizer.is_valid())

    async def test_refresh__with_invalid_token(self):
        authorizer = asyncprawcore.Authorizer(
            self.authentication, refresh_token="INVALID_TOKEN"
        )
        with VCR.use_cassette("Authorizer_refresh__with_invalid_token"):
            with self.assertRaises(asyncprawcore.ResponseException):
                await authorizer.refresh()
            self.assertFalse(authorizer.is_valid())

    async def test_refresh__without_refresh_token(self):
        authorizer = asyncprawcore.Authorizer(self.authentication)
        with self.assertRaises(asyncprawcore.InvalidInvocation):
            await authorizer.refresh()
        self.assertFalse(authorizer.is_valid())

    async def test_revoke__access_token_with_refresh_set(self):
        authorizer = asyncprawcore.Authorizer(
            self.authentication, refresh_token=REFRESH_TOKEN
        )
        with VCR.use_cassette("Authorizer_revoke__access_token_with_refresh_set"):
            await authorizer.refresh()
            await authorizer.revoke(only_access=True)

            self.assertIsNone(authorizer.access_token)
            self.assertIsNotNone(authorizer.refresh_token)
            self.assertIsNone(authorizer.scopes)
            self.assertFalse(authorizer.is_valid())

            await authorizer.refresh()

        self.assertTrue(authorizer.is_valid())

    async def test_revoke__access_token_without_refresh_set(self):
        self.authentication.redirect_uri = REDIRECT_URI
        authorizer = asyncprawcore.Authorizer(self.authentication)
        with VCR.use_cassette("Authorizer_revoke__access_token_without_refresh_set"):
            await authorizer.authorize(TEMPORARY_GRANT_CODE)
            await authorizer.revoke()

        self.assertIsNone(authorizer.access_token)
        self.assertIsNone(authorizer.refresh_token)
        self.assertIsNone(authorizer.scopes)
        self.assertFalse(authorizer.is_valid())

    async def test_revoke__refresh_token_with_access_set(self):
        authorizer = asyncprawcore.Authorizer(
            self.authentication, refresh_token=REFRESH_TOKEN
        )
        with VCR.use_cassette("Authorizer_revoke__refresh_token_with_access_set"):
            await authorizer.refresh()
            await authorizer.revoke()

        self.assertIsNone(authorizer.access_token)
        self.assertIsNone(authorizer.refresh_token)
        self.assertIsNone(authorizer.scopes)
        self.assertFalse(authorizer.is_valid())

    async def test_revoke__refresh_token_without_access_set(self):
        authorizer = asyncprawcore.Authorizer(
            self.authentication, refresh_token=REFRESH_TOKEN
        )
        with VCR.use_cassette("Authorizer_revoke__refresh_token_without_access_set"):
            await authorizer.revoke()

        self.assertIsNone(authorizer.access_token)
        self.assertIsNone(authorizer.refresh_token)
        self.assertIsNone(authorizer.scopes)
        self.assertFalse(authorizer.is_valid())

    async def test_revoke__without_access_token(self):
        authorizer = asyncprawcore.Authorizer(
            self.authentication, refresh_token=REFRESH_TOKEN
        )
        with self.assertRaises(asyncprawcore.InvalidInvocation):
            await authorizer.revoke(only_access=True)

    async def test_revoke__without_any_token(self):
        authorizer = asyncprawcore.Authorizer(self.authentication)
        with self.assertRaises(asyncprawcore.InvalidInvocation):
            await authorizer.revoke()


class DeviceIDAuthorizerTest(AuthorizerTestBase):
    async def setUp(self):
        self.requestor = Requestor("asyncprawcore:test (by /u/bboe)")

        self.authentication = asyncprawcore.UntrustedAuthenticator(
            self.requestor, CLIENT_ID
        )

    async def test_initialize(self):
        authorizer = asyncprawcore.DeviceIDAuthorizer(self.authentication)
        self.assertIsNone(authorizer.access_token)
        self.assertIsNone(authorizer.scopes)
        self.assertFalse(authorizer.is_valid())

    async def test_initialize__with_trusted_authenticator(self):
        authenticator = asyncprawcore.TrustedAuthenticator(None, None, None)
        with self.assertRaises(asyncprawcore.InvalidInvocation):
            asyncprawcore.DeviceIDAuthorizer(authenticator)

    async def test_refresh(self):
        authorizer = asyncprawcore.DeviceIDAuthorizer(self.authentication)
        with VCR.use_cassette("DeviceIDAuthorizer_refresh"):
            await authorizer.refresh()

        self.assertIsNotNone(authorizer.access_token)
        self.assertEqual(set(["*"]), authorizer.scopes)
        self.assertTrue(authorizer.is_valid())

    async def test_refresh__with_short_device_id(self):
        authorizer = asyncprawcore.DeviceIDAuthorizer(self.authentication, "a" * 19)
        with VCR.use_cassette("DeviceIDAuthorizer_refresh__with_short_device_id"):
            with self.assertRaises(asyncprawcore.OAuthException):
                await authorizer.refresh()


class ImplicitAuthorizerTest(AuthorizerTestBase):
    def test_initialize(self):
        authenticator = asyncprawcore.UntrustedAuthenticator(self.requestor, CLIENT_ID)
        authorizer = asyncprawcore.ImplicitAuthorizer(
            authenticator, "fake token", 1, "modposts read"
        )
        self.assertEqual("fake token", authorizer.access_token)
        self.assertEqual({"modposts", "read"}, authorizer.scopes)
        self.assertTrue(authorizer.is_valid())

    def test_initialize__with_trusted_authenticator(self):
        self.assertRaises(
            asyncprawcore.InvalidInvocation,
            asyncprawcore.ImplicitAuthorizer,
            self.authentication,
            None,
            None,
            None,
        )


class ReadOnlyAuthorizerTest(AuthorizerTestBase):
    def test_initialize__with_untrusted_authenticator(self):
        authenticator = asyncprawcore.UntrustedAuthenticator(self.requestor, CLIENT_ID)
        self.assertRaises(
            asyncprawcore.InvalidInvocation,
            asyncprawcore.ReadOnlyAuthorizer,
            authenticator,
        )

    async def test_refresh(self):
        authorizer = asyncprawcore.ReadOnlyAuthorizer(self.authentication)
        self.assertIsNone(authorizer.access_token)
        self.assertIsNone(authorizer.scopes)
        self.assertFalse(authorizer.is_valid())

        with VCR.use_cassette("ReadOnlyAuthorizer_refresh"):
            await authorizer.refresh()

        self.assertIsNotNone(authorizer.access_token)
        self.assertEqual(set(["*"]), authorizer.scopes)
        self.assertTrue(authorizer.is_valid())


class ScriptAuthorizerTest(AuthorizerTestBase):
    def test_initialize__with_untrusted_authenticator(self):
        authenticator = asyncprawcore.UntrustedAuthenticator(self.requestor, CLIENT_ID)
        self.assertRaises(
            asyncprawcore.InvalidInvocation,
            asyncprawcore.ScriptAuthorizer,
            authenticator,
            None,
            None,
        )

    async def test_refresh(self):
        authorizer = asyncprawcore.ScriptAuthorizer(
            self.authentication, USERNAME, PASSWORD
        )
        self.assertIsNone(authorizer.access_token)
        self.assertIsNone(authorizer.scopes)
        self.assertFalse(authorizer.is_valid())

        with VCR.use_cassette("ScriptAuthorizer_refresh"):
            await authorizer.refresh()

        self.assertIsNotNone(authorizer.access_token)
        self.assertEqual({"*"}, authorizer.scopes)
        self.assertTrue(authorizer.is_valid())

    async def test_refresh__with_invalid_username_or_password(self):
        authorizer = asyncprawcore.ScriptAuthorizer(
            self.authentication, USERNAME, "invalidpassword"
        )
        with VCR.use_cassette(
            "ScriptAuthorizer_refresh__with_invalid_username_or_password"
        ):
            with self.assertRaises(asyncprawcore.OAuthException):
                await authorizer.refresh()
            self.assertFalse(authorizer.is_valid())



"""Constants for the asyncprawcore package."""
import os

__version__ = "2.0.0"

ACCESS_TOKEN_PATH = "/api/v1/access_token"
AUTHORIZATION_PATH = "/api/v1/authorize"
REVOKE_TOKEN_PATH = "/api/v1/revoke_token"
TIMEOUT = float(os.environ.get("prawcore_timeout", 16))



"""Test for sublcasses of asyncprawcore.auth.BaseAuthenticator class."""
import asyncprawcore
import asynctest

from .conftest import CLIENT_ID, CLIENT_SECRET, REDIRECT_URI, VCR
from asyncprawcore.requestor import Requestor


class TrustedAuthenticatorTest(asynctest.TestCase):
    async def setUp(self) -> None:
        self.requestor = Requestor("asyncprawcore:test (by /u/bboe)")

    async def tearDown(self) -> None:
        await self.requestor.close()

    async def test_authorize_url(self):
        authenticator = asyncprawcore.TrustedAuthenticator(
            self.requestor, CLIENT_ID, CLIENT_SECRET, REDIRECT_URI
        )
        url = authenticator.authorize_url("permanent", ["identity", "read"], "a_state")
        self.assertIn(f"client_id={CLIENT_ID}", url)
        self.assertIn("duration=permanent", url)
        self.assertIn("response_type=code", url)
        self.assertIn("scope=identity+read", url)
        self.assertIn("state=a_state", url)

    def test_authorize_url__fail_with_implicit(self):
        authenticator = asyncprawcore.TrustedAuthenticator(
            self.requestor, CLIENT_ID, CLIENT_SECRET, REDIRECT_URI
        )
        self.assertRaises(
            asyncprawcore.InvalidInvocation,
            authenticator.authorize_url,
            "temporary",
            ["identity", "read"],
            "a_state",
            implicit=True,
        )

    def test_authorize_url__fail_without_redirect_uri(self):
        authenticator = asyncprawcore.TrustedAuthenticator(
            self.requestor, CLIENT_ID, CLIENT_SECRET
        )
        self.assertRaises(
            asyncprawcore.InvalidInvocation,
            authenticator.authorize_url,
            "permanent",
            ["identity"],
            "...",
        )

    async def test_revoke_token(self):
        authenticator = asyncprawcore.TrustedAuthenticator(
            self.requestor, CLIENT_ID, CLIENT_SECRET
        )
        with VCR.use_cassette("TrustedAuthenticator_revoke_token"):
            await authenticator.revoke_token("dummy token")

    async def test_revoke_token__with_access_token_hint(self):
        authenticator = asyncprawcore.TrustedAuthenticator(
            self.requestor, CLIENT_ID, CLIENT_SECRET
        )
        with VCR.use_cassette(
            "TrustedAuthenticator_revoke_token__with_access_token_hint"
        ):
            await authenticator.revoke_token("dummy token", "access_token")

    async def test_revoke_token__with_refresh_token_hint(self):
        authenticator = asyncprawcore.TrustedAuthenticator(
            self.requestor, CLIENT_ID, CLIENT_SECRET
        )
        with VCR.use_cassette(
            "TrustedAuthenticator_revoke_token__with_refresh_token_hint"
        ):
            await authenticator.revoke_token("dummy token", "refresh_token")


class UntrustedAuthenticatorTest(asynctest.TestCase):
    async def setUp(self) -> None:
        self.requestor = Requestor("asyncprawcore:test (by /u/bboe)")

    async def tearDown(self) -> None:
        await self.requestor.close()

    async def test_authorize_url__code(self):
        authenticator = asyncprawcore.UntrustedAuthenticator(
            self.requestor, CLIENT_ID, REDIRECT_URI
        )
        url = authenticator.authorize_url("permanent", ["identity", "read"], "a_state")
        self.assertIn(f"client_id={CLIENT_ID}", url)
        self.assertIn("duration=permanent", url)
        self.assertIn("response_type=code", url)
        self.assertIn("scope=identity+read", url)
        self.assertIn("state=a_state", url)

    async def test_authorize_url__token(self):
        authenticator = asyncprawcore.UntrustedAuthenticator(
            self.requestor, CLIENT_ID, REDIRECT_URI
        )
        url = authenticator.authorize_url(
            "temporary", ["identity", "read"], "a_state", implicit=True
        )
        self.assertIn(f"client_id={CLIENT_ID}", url)
        self.assertIn("duration=temporary", url)
        self.assertIn("response_type=token", url)
        self.assertIn("scope=identity+read", url)
        self.assertIn("state=a_state", url)

    async def test_authorize_url__fail_with_token_and_permanent(self):
        authenticator = asyncprawcore.UntrustedAuthenticator(
            self.requestor, CLIENT_ID, REDIRECT_URI
        )
        with self.assertRaises(asyncprawcore.InvalidInvocation):
            authenticator.authorize_url(
                "permanent",
                ["identity", "read"],
                "a_state",
                implicit=True,
            )

    def test_authorize_url__fail_without_redirect_uri(self):
        authenticator = asyncprawcore.UntrustedAuthenticator(self.requestor, CLIENT_ID)
        self.assertRaises(
            asyncprawcore.InvalidInvocation,
            authenticator.authorize_url,
            "temporary",
            ["identity"],
            "...",
        )

    async def test_revoke_token(self):
        authenticator = asyncprawcore.UntrustedAuthenticator(self.requestor, CLIENT_ID)
        with VCR.use_cassette("UntrustedAuthenticator_revoke_token"):
            await authenticator.revoke_token("dummy token")



"""Constants for the asyncprawcore test suite."""

import json
import os
from base64 import b64encode
from datetime import datetime
from vcr import VCR
from vcr.persisters.filesystem import FilesystemPersister
from vcr.serialize import deserialize, serialize


CLIENT_ID = os.environ.get("PRAWCORE_CLIENT_ID", "fake_client_id")
CLIENT_SECRET = os.environ.get("PRAWCORE_CLIENT_SECRET", "fake_client_secret")
PASSWORD = os.environ.get("PRAWCORE_PASSWORD", "fake_password")
PERMANENT_GRANT_CODE = os.environ.get("PRAWCORE_PERMANENT_GRANT_CODE", "fake_perm_code")
REDIRECT_URI = os.environ.get("PRAWCORE_REDIRECT_URI", "http://localhost:8080")
REFRESH_TOKEN = os.environ.get("PRAWCORE_REFRESH_TOKEN", "fake_refresh_token")
TEMPORARY_GRANT_CODE = os.environ.get("PRAWCORE_TEMPORARY_GRANT_CODE", "fake_temp_code")
USERNAME = os.environ.get("PRAWCORE_USERNAME", "fake_username")


def b64_string(input_string):
    """Return a base64 encoded string (not bytes) from input_string."""
    return b64encode(input_string.encode("utf-8")).decode("utf-8")


placeholders = [
    ("<CLIENT_ID>", CLIENT_ID),
    ("<CLIENT_SECRET>", CLIENT_SECRET),
    ("<PASSWORD>", PASSWORD),
    ("<PERM_CODE>", PERMANENT_GRANT_CODE),
    ("<REFRESH_TOKEN>", REFRESH_TOKEN),
    ("<TEMP_CODE>", TEMPORARY_GRANT_CODE),
    ("<USERNAME>", USERNAME),
    ("<BASIC_AUTH>", b64_string(f"{CLIENT_ID}:{CLIENT_SECRET}")),
]


def filter_access_token(response):
    """Add VCR callback to filter access token."""
    request_uri = response["url"]
    if "api/v1/access_token" not in request_uri or response["status"]["code"] != 200:
        return response
    body = response["body"]["string"].decode()
    try:
        token = json.loads(body)["access_token"]
        response["body"]["string"] = response["body"]["string"].replace(
            token.encode("utf-8"), b"<ACCESS_TOKEN>"
        )
        placeholders["access_token"] = token
    except (KeyError, TypeError, ValueError):
        pass
    return response


def serialize_list(data: list):
    """List serializer."""
    new_list = []
    for item in data:
        if isinstance(item, dict):
            new_list.append(serialize_dict(item))
        elif isinstance(item, list):
            new_list.append(serialize_list(item))
        else:
            new_list.append(item)
    return new_list


def serialize_dict(data: dict):
    """Filter out buffered readers."""
    new_dict = {}
    for key, value in data.items():
        if key == "file":
            continue  # skip files
        elif isinstance(value, dict):
            new_dict[key] = serialize_dict(value)
        elif isinstance(value, list):
            new_dict[key] = serialize_list(value)
        else:
            new_dict[key] = value
    return new_dict


class CustomSerializer(object):
    """Custom serializer for cassettes."""

    @staticmethod
    def serialize(cassette_dict):
        """Serialize cassette dict."""
        cassette_dict["recorded_at"] = datetime.now().isoformat()[:-7]
        return (
            f"{json.dumps(serialize_dict(cassette_dict), sort_keys=True, indent=2)}\n"
        )

    @staticmethod
    def deserialize(cassette_string):
        """Deserialize cassette string."""
        return json.loads(cassette_string)


class CustomPersister(FilesystemPersister):
    """Custom persiter for VCR."""

    @classmethod
    def load_cassette(cls, cassette_path, serializer):
        """Load the cassette."""
        try:
            with open(cassette_path) as f:
                cassette_content = f.read()
        except OSError:
            raise ValueError("Cassette not found.")
        for replacement, value in placeholders:
            cassette_content = cassette_content.replace(value, replacement)
        cassette = deserialize(cassette_content, serializer)
        return cassette

    @staticmethod
    def save_cassette(cassette_path, cassette_dict, serializer):
        """Save the cassette."""
        data = serialize(cassette_dict, serializer)
        for replacement, value in placeholders:
            data = data.replace(value, replacement)
        dirname, filename = os.path.split(cassette_path)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
        with open(cassette_path, "w") as f:
            f.write(data)


class CustomVCR(VCR):
    """Derived from VCR to make setting paths easier."""

    def use_cassette(self, path="", **kwargs):
        """Use a cassette."""
        path += ".json"
        return super().use_cassette(path, **kwargs)


VCR = CustomVCR(
    serializer="custom_serializer",
    cassette_library_dir="tests/cassettes",
    match_on=["uri", "method"],
    before_record_response=filter_access_token,
)
VCR.register_serializer("custom_serializer", CustomSerializer)
VCR.register_persister(CustomPersister)


class AsyncMock:
    """Class to assist making asynchronous mocks simpler to write."""

    def __init__(self, status, response_dict, headers):
        """Initialize the class with return status, response-dict and headers."""
        self.status = status
        self.response_dict = response_dict
        self.headers = headers

    async def json(self):
        """Mock the json of ClientSession.request."""
        return self.response_dict



"""Test for asyncprawcore.Sessions module."""
import asyncio
import logging
from json import dumps

import asynctest
from aiohttp.web import HTTPRequestTimeout
from mock import patch
from testfixtures import LogCapture

import asyncprawcore
from asyncprawcore.exceptions import RequestException
from .conftest import (
    AsyncMock,
    CLIENT_ID,
    CLIENT_SECRET,
    PASSWORD,
    REFRESH_TOKEN,
    USERNAME,
    VCR,
)


class InvalidAuthorizer(asyncprawcore.Authorizer):
    def __init__(self):
        requestor = asyncprawcore.requestor.Requestor("asyncprawcore:test (by /u/bboe)")

        super(InvalidAuthorizer, self).__init__(
            asyncprawcore.TrustedAuthenticator(requestor, CLIENT_ID, CLIENT_SECRET)
        )

    def is_valid(self):
        return False


async def client_authorizer():
    requestor = asyncprawcore.requestor.Requestor("asyncprawcore:test (by /u/bboe)")

    authenticator = asyncprawcore.TrustedAuthenticator(
        requestor, CLIENT_ID, CLIENT_SECRET
    )
    authorizer = asyncprawcore.Authorizer(authenticator, refresh_token=REFRESH_TOKEN)
    await authorizer.refresh()
    return authorizer


async def readonly_authorizer(refresh=True):
    requestor = asyncprawcore.requestor.Requestor("asyncprawcore:test (by /u/bboe)")
    authenticator = asyncprawcore.TrustedAuthenticator(
        requestor, CLIENT_ID, CLIENT_SECRET
    )
    authorizer = asyncprawcore.ReadOnlyAuthorizer(authenticator)
    if refresh:
        await authorizer.refresh()
    return authorizer


async def script_authorizer():
    requestor = asyncprawcore.requestor.Requestor("asyncprawcore:test (by /u/bboe)")
    authenticator = asyncprawcore.TrustedAuthenticator(
        requestor, CLIENT_ID, CLIENT_SECRET
    )
    authorizer = asyncprawcore.ScriptAuthorizer(authenticator, USERNAME, PASSWORD)
    await authorizer.refresh()
    return authorizer


class SessionTest(asynctest.TestCase):
    async def setUp(self) -> None:
        self.requestor = asyncprawcore.requestor.Requestor(
            "asyncprawcore:test (by /u/bboe)"
        )

    async def tearDown(self) -> None:
        await self.requestor.close()
        if hasattr(self, "session"):
            await self.session.close()

    async def test_close(self):
        authorizer = await readonly_authorizer(refresh=False)
        await asyncprawcore.Session(authorizer).close()

    async def test_context_manager(self):
        async with asyncprawcore.Session(
            await readonly_authorizer(refresh=False)
        ) as session:
            self.assertIsInstance(session, asyncprawcore.Session)

    def test_init__without_authenticator(self):
        self.assertRaises(asyncprawcore.InvalidInvocation, asyncprawcore.Session, None)

    def test_init__with_device_id_authorizer(self):
        authenticator = asyncprawcore.UntrustedAuthenticator(self.requestor, CLIENT_ID)
        authorizer = asyncprawcore.DeviceIDAuthorizer(authenticator)
        asyncprawcore.Session(authorizer)

    def test_init__with_implicit_authorizer(self):
        authenticator = asyncprawcore.UntrustedAuthenticator(self.requestor, CLIENT_ID)
        authorizer = asyncprawcore.ImplicitAuthorizer(authenticator, None, 0, "")
        asyncprawcore.Session(authorizer)

    @patch("aiohttp.ClientSession")
    async def test_request__connection_error_retry(self, mock_session):
        session_instance = mock_session.return_value
        try:
            session_instance.request.return_value = asyncio.Future()
            session_instance.request.return_value.set_result(
                AsyncMock(
                    status=200,
                    response_dict={
                        "access_token": "",
                        "expires_in": 99,
                        "scope": "",
                    },
                    headers={},
                )
            )

            authorizer = await readonly_authorizer()
            session_instance.request.reset_mock()

            # Fail on subsequent request
            exception = ConnectionError()
            session_instance.request.side_effect = exception

            expected = (
                "asyncprawcore",
                "WARNING",
                "Retrying due to ConnectionError() status: GET "
                "https://oauth.reddit.com/",
            )
        finally:
            session_instance.close()

        with LogCapture(level=logging.WARNING) as log_capture:
            with self.assertRaises(RequestException) as context_manager:
                await asyncprawcore.Session(authorizer).request("GET", "/")
            log_capture.check(expected, expected)
        self.assertIsInstance(context_manager.exception, RequestException)
        self.assertIs(exception, context_manager.exception.original_exception)
        self.assertEqual(3, session_instance.request.call_count)

    async def test_request__get(self):
        with VCR.use_cassette("Session_request__get"):
            self.session = asyncprawcore.Session(await readonly_authorizer())
            params = {"limit": 100, "bool_param": True}
            response = await self.session.request("GET", "/", params=params)
        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(params))
        self.assertEqual("Listing", response["kind"])

    async def test_request__patch(self):
        with VCR.use_cassette(
            "Session_request__patch",
            match_requests_on=["method", "uri", "json-body"],
        ):
            self.session = asyncprawcore.Session(await script_authorizer())
            json = {"lang": "ja", "num_comments": 123}
            response = await self.session.request(
                "PATCH", "/api/v1/me/prefs", json=json
            )
            self.assertEqual("ja", response["lang"])
            self.assertEqual(123, response["num_comments"])

    async def test_request__post(self):
        with VCR.use_cassette("Session_request__post"):
            session = asyncprawcore.Session(await script_authorizer())
            data = {
                "kind": "self",
                "sr": "asyncpraw",
                "text": "Test!",
                "title": "A Test from asyncprawcore.",
            }
            key_count = len(data)
            response = await session.request("POST", "/api/submit", data=data)
            self.assertIn("a_test_from_asyncprawcore", response["json"]["data"]["url"])
            self.assertEqual(key_count, len(data))  # Ensure data is untouched

    async def test_request__post__with_files(self):
        with VCR.use_cassette("Session_request__post__with_files", match_on=["uri"]):
            session = asyncprawcore.Session(await script_authorizer())
            with open("./tests/files/white-square.png", "rb") as fp:
                files = {"file": fp}
                data = {"test": "data"}
                response = await session.request(
                    "POST", "/r/asyncpraw/api/upload_sr_img", files=files, data=data
                )
            self.assertIn("img_src", response)

    async def test_request__raw_json(self):
        with VCR.use_cassette("Session_request__raw_json"):
            self.session = asyncprawcore.Session(await readonly_authorizer())
            response = await self.session.request(
                "GET",
                "/r/reddit_api_test/comments/45xjdr/want_raw_json_test/",
            )
        self.assertEqual(
            "WANT_RAW_JSON test: < > &",
            response[0]["data"]["children"][0]["data"]["title"],
        )

    async def test_request__bad_gateway(self):
        with VCR.use_cassette("Session_request__bad_gateway"):
            self.session = asyncprawcore.Session(await readonly_authorizer())
            with self.assertRaises(asyncprawcore.ServerError) as context_manager:
                await self.session.request("GET", "/")
            self.assertEqual(502, context_manager.exception.response.status)

    async def test_request__bad_json(self):
        with VCR.use_cassette("Session_request__bad_json"):
            self.session = asyncprawcore.Session(await script_authorizer())
            with self.assertRaises(asyncprawcore.BadJSON) as context_manager:
                await self.session.request("GET", "/")
            self.assertEqual(17512, context_manager.exception.response.content_length)

    async def test_request__bad_request(self):
        with VCR.use_cassette("Session_request__bad_request"):
            self.session = asyncprawcore.Session(await script_authorizer())
            with self.assertRaises(asyncprawcore.BadRequest) as context_manager:
                await self.session.request(
                    "PUT",
                    "/api/v1/me/friends/spez",
                    data='{"note": "asyncprawcore"}',
                )
            self.assertIn("reason", await context_manager.exception.response.json())

    async def test_request__cloudflair_connection_timed_out(self):
        with VCR.use_cassette("Session_request__cloudflair_connection_timed_out"):
            self.session = asyncprawcore.Session(await readonly_authorizer())
            with self.assertRaises(asyncprawcore.ServerError) as context_manager:
                await self.session.request("GET", "/")
                await self.session.request("GET", "/")
                await self.session.request("GET", "/")
            self.assertEqual(522, context_manager.exception.response.status)

    async def test_request__cloudflair_unknown_error(self):
        with VCR.use_cassette("Session_request__cloudflair_unknown_error"):
            self.session = asyncprawcore.Session(await readonly_authorizer())
            with self.assertRaises(asyncprawcore.ServerError) as context_manager:
                await self.session.request("GET", "/")
                await self.session.request("GET", "/")
                await self.session.request("GET", "/")
            self.assertEqual(520, context_manager.exception.response.status)

    async def test_request__conflict(self):
        with VCR.use_cassette("Session_request__conflict"):
            session = asyncprawcore.Session(await script_authorizer())
            with self.assertRaises(asyncprawcore.Conflict) as context_manager:
                await session.request(
                    "POST",
                    "/api/multi/copy/",
                    data={
                        "display_name": "sfwpornnetwork",
                        "from": "/user/kjoneslol/m/sfwpornnetwork",
                        "to": f"user/{USERNAME}/m/sfwpornnetwork/",
                    },
                )
            self.assertEqual(409, context_manager.exception.response.status)

    async def test_request__created(self):
        with VCR.use_cassette("Session_request__created"):
            self.session = asyncprawcore.Session(await script_authorizer())
            response = await self.session.request(
                "PUT", "/api/v1/me/friends/spez", data="{}"
            )
            self.assertIn("name", response)

    async def test_request__forbidden(self):
        with VCR.use_cassette("Session_request__forbidden"):
            self.session = asyncprawcore.Session(await script_authorizer())
            with self.assertRaises(asyncprawcore.Forbidden):
                await self.session.request("GET", "/user/spez/gilded/given")

    async def test_request__gateway_timeout(self):
        with VCR.use_cassette("Session_request__gateway_timeout"):
            self.session = asyncprawcore.Session(await readonly_authorizer())
            with self.assertRaises(asyncprawcore.ServerError) as context_manager:
                await self.session.request("GET", "/")
            self.assertEqual(504, context_manager.exception.response.status)

    async def test_request__internal_server_error(self):
        with VCR.use_cassette("Session_request__internal_server_error"):
            self.session = asyncprawcore.Session(await readonly_authorizer())
            with self.assertRaises(asyncprawcore.ServerError) as context_manager:
                await self.session.request("GET", "/")
            self.assertEqual(500, context_manager.exception.response.status)

    async def test_request__no_content(self):
        with VCR.use_cassette("Session_request__no_content"):
            self.session = asyncprawcore.Session(await script_authorizer())
            response = await self.session.request("DELETE", "/api/v1/me/friends/spez")
            self.assertIsNone(response)

    async def test_request__not_found(self):
        with VCR.use_cassette("Session_request__not_found"):
            self.session = asyncprawcore.Session(await script_authorizer())
            with self.assertRaises(asyncprawcore.NotFound):
                await self.session.request("GET", "/r/cricket/wiki/invalid")

    async def test_request__okay_with_0_byte_content(self):
        with VCR.use_cassette(
            "Session_request__okay_with_0_byte_content", match_on=["method"]
        ):
            self.session = asyncprawcore.Session(await script_authorizer())
            data = {"model": dumps({"name": "test"})}
            path = f"/api/multi/user/{USERNAME}/m/test"
            response = await self.session.request("DELETE", path, data=data)
            self.assertEqual("", response)

    @patch("aiohttp.ClientSession")
    async def test_request__read_timeout_retry(self, mock_session):
        self.session_instance = mock_session.return_value
        self.session_instance.request.return_value = asyncio.Future()
        self.session_instance.request.return_value.set_result(
            AsyncMock(
                status=200,
                response_dict={
                    "access_token": "",
                    "expires_in": 99,
                    "scope": "",
                },
                headers={},
            )
        )

        authorizer = await readonly_authorizer()
        self.session_instance.request.reset_mock()

        exception = HTTPRequestTimeout()
        self.session_instance.request.side_effect = exception

        expected = (
            "asyncprawcore",
            "WARNING",
            "Retrying due to <HTTPRequestTimeout Request Timeout not prepared> status: "
            "GET https://oauth.reddit.com/",
        )

        with LogCapture(level=logging.WARNING) as log_capture:
            with self.assertRaises(RequestException) as context_manager:
                await asyncprawcore.Session(authorizer).request("GET", "/")
            log_capture.check(expected, expected)
        self.assertIsInstance(context_manager.exception, RequestException)
        self.assertIs(exception, context_manager.exception.original_exception)
        self.assertEqual(3, self.session_instance.request.call_count)

    async def test_request__redirect(self):
        with VCR.use_cassette("Session_request__redirect"):
            session = asyncprawcore.Session(await readonly_authorizer())
            with self.assertRaises(asyncprawcore.Redirect) as context_manager:
                await session.request("GET", "/r/random")
            self.assertTrue(context_manager.exception.path.startswith("/r/"))

    async def test_request__service_unavailable(self):
        with VCR.use_cassette("Session_request__service_unavailable"):
            self.session = asyncprawcore.Session(await readonly_authorizer())
            with self.assertRaises(asyncprawcore.ServerError) as context_manager:
                await self.session.request("GET", "/")
                await self.session.request("GET", "/")
                await self.session.request("GET", "/")
            self.assertEqual(503, context_manager.exception.response.status)

    async def test_request__too_large(self):
        with VCR.use_cassette(
            "Session_request__too_large",
            match_requests_on=["uri", "method"],  # , serializer="yaml",
        ):
            session = asyncprawcore.Session(await script_authorizer())
            with self.assertRaises(asyncprawcore.TooLarge) as context_manager:
                await session.request(
                    "POST",
                    "/r/asyncpraw/api/upload_sr_img",
                    files={"file": open("./tests/files/too_large.jpg", "rb")},
                )
            self.assertEqual(413, context_manager.exception.response.status)

    async def test_request__unavailable_for_legal_reasons(self):
        with VCR.use_cassette("Session_request__unavailable_for_legal_reasons"):
            authenticator = asyncprawcore.UntrustedAuthenticator(
                self.requestor, CLIENT_ID
            )
            authorizer = asyncprawcore.ImplicitAuthorizer(authenticator, None, 0, "")
            self.session = asyncprawcore.Session(authorizer)
            exception_class = asyncprawcore.UnavailableForLegalReasons
            with self.assertRaises(exception_class) as context_manager:
                await self.session.request("GET", "/")
            self.assertEqual(451, context_manager.exception.response.status)

    async def test_request__unsupported_media_type(self):
        with VCR.use_cassette(
            "Session_request__unsupported_media_type",
            match_requests_on=["uri", "method"],
        ):
            session = asyncprawcore.Session(await script_authorizer())
            exception_class = asyncprawcore.SpecialError
            data = {
                "content": "type: submission\naction: upvote",
                "page": "config/automoderator",
            }
            with self.assertRaises(exception_class) as context_manager:
                await session.request("POST", "r/asyncpraw/api/wiki/edit/", data=data)
            self.assertEqual(415, context_manager.exception.response.status)

    async def test_request__with_insufficient_scope(self):
        with VCR.use_cassette("Session_request__with_insufficient_scope"):
            self.session = asyncprawcore.Session(await client_authorizer())
            with self.assertRaises(asyncprawcore.InsufficientScope):
                await self.session.request("GET", "/api/v1/me")

    async def test_request__with_invalid_access_token(self):
        self.authenticator = asyncprawcore.UntrustedAuthenticator(
            self.requestor, CLIENT_ID
        )
        self.authorizer = asyncprawcore.ImplicitAuthorizer(
            self.authenticator, None, 0, ""
        )
        self.session = asyncprawcore.Session(self.authorizer)

        with VCR.use_cassette("Session_request__with_invalid_access_token"):
            self.session._authorizer.access_token = "invalid"
            with self.assertRaises(asyncprawcore.InvalidToken):
                await self.session.request("get", "/")

    async def test_request__with_invalid_access_token__retry(self):
        with VCR.use_cassette("Session_request__with_invalid_access_token__retry"):
            self.session = asyncprawcore.Session(await readonly_authorizer())
            self.session._authorizer.access_token += "invalid"
            response = await self.session.request("GET", "/")
        self.assertIsInstance(response, dict)

    async def test_request__with_invalid_authorizer(self):
        self.session = asyncprawcore.Session(InvalidAuthorizer())
        with self.assertRaises(asyncprawcore.InvalidInvocation):
            await self.session.request("get", "/")


class SessionFunctionTest(asynctest.TestCase):
    async def test_session(self):
        session = asyncprawcore.session(InvalidAuthorizer())
        try:
            self.assertIsInstance(session, asyncprawcore.Session)
        finally:
            await session.close()



"""asyncprawcore setup.py."""

from codecs import open
from os import path
from setuptools import setup


PACKAGE_NAME = "asyncprawcore"
HERE = path.abspath(path.dirname(__file__))
with open(path.join(HERE, "README.rst"), encoding="utf-8") as fp:
    README = fp.read()
with open(path.join(HERE, PACKAGE_NAME, "const.py"), encoding="utf-8") as fp:
    VERSION_LINE = next(line for line in fp.readlines() if "__version__" in line)
    VERSION = VERSION_LINE.split("=")[-1].strip().replace('"', "")

extras = {
    "ci": ["coveralls"],
    "lint": ["black", "flake8", "pre-commit", "pydocstyle", "flynt"],
    "test": [
        "pytest-vcr",
        "vcrpy==4.0.2",
        "mock >=0.8",
        "pytest",
        "testfixtures >4.13.2, <7",
        "asynctest >=0.13.0",
    ],
}
extras["dev"] = extras["lint"] + extras["test"]

setup(
    name=PACKAGE_NAME,
    author="LilSpazJoekp,vikramaditya91",
    author_email="lilspazjoekp@gmail.com,vikramaditya91@gmail.com",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    description="Low-level asynchronous communication layer for Async PRAW 7+.",
    extras_require=extras,
    install_requires=["aiohttp", "yarl"],
    python_requires=">=3.6",
    keywords="praw reddit api async",
    license="Simplified BSD License",
    long_description=README,
    packages=[PACKAGE_NAME],
    url="https://github.com/praw-dev/asyncprawcore",
    version=VERSION,
)



"""Provides Authentication and Authorization classes."""
import aiohttp
import inspect
import time

from aiohttp import ClientRequest
from yarl import URL

from . import const
from .codes import codes
from .exceptions import InvalidInvocation, OAuthException, ResponseException


class BaseAuthenticator(object):
    """Provide the base authenticator object that stores OAuth2 credentials."""

    def __init__(self, requestor, client_id, redirect_uri=None):
        """Represent a single authentication to Reddit's API.

        :param requestor: An instance of :class:`Requestor`.
        :param client_id: The OAuth2 client ID to use with the session.
        :param redirect_uri: (optional) The redirect URI exactly as specified
            in your OAuth application settings on Reddit. This parameter is
            required if you want to use the ``authorize_url`` method, or the
            ``authorize`` method of the ``Authorizer`` class.

        """
        self._requestor = requestor
        self.client_id = client_id
        self.redirect_uri = redirect_uri

    async def _post(self, url, success_status=codes["ok"], **data):
        response = await self._requestor.request(
            "POST",
            url,
            auth=self._auth(),
            data=sorted(data.items()),
            headers={"Connection": "close"},
        )
        if response.status != success_status:
            raise ResponseException(response)
        return response

    def authorize_url(self, duration, scopes, state, implicit=False):
        """Return the URL used out-of-band to grant access to your application.

        :param duration: Either ``permanent`` or ``temporary``. ``temporary``
            authorizations generate access tokens that last only 1
            hour. ``permanent`` authorizations additionally generate a refresh
            token that can be indefinitely used to generate new hour-long
            access tokens. Only ``temporary`` can be specified if ``implicit``
            is set to ``True``.
        :param scopes: A list of OAuth scopes to request authorization for.
        :param state: A string that will be reflected in the callback to
            ``redirect_uri``. Elements must be printable ASCII characters in the range
            0x20 through 0x7E inclusive. This value should be temporarily unique to the
            client for whom the URL was generated.
        :param implicit: (optional) Use the implicit grant flow (default:
            False). This flow is only available for UntrustedAuthenticators.

        """
        if self.redirect_uri is None:
            raise InvalidInvocation("redirect URI not provided")
        if implicit and not isinstance(self, UntrustedAuthenticator):
            raise InvalidInvocation(
                "Only UntrustedAuthenticator instances can "
                "use the implicit grant flow."
            )
        if implicit and duration != "temporary":
            raise InvalidInvocation(
                "The implicit grant flow only supports temporary access tokens."
            )

        params = {
            "client_id": self.client_id,
            "duration": duration,
            "redirect_uri": self.redirect_uri,
            "response_type": "token" if implicit else "code",
            "scope": " ".join(scopes),
            "state": state,
        }
        url = self._requestor.reddit_url + const.AUTHORIZATION_PATH
        request = ClientRequest("GET", URL(url), params=params)
        return str(request.url)

    async def revoke_token(self, token, token_type=None):
        """Ask Reddit to revoke the provided token.

        :param token: The access or refresh token to revoke.
        :param token_type: (Optional) When provided, hint to Reddit what the
            token type is for a possible efficiency gain. The value can be
            either ``access_token`` or ``refresh_token``.

        """
        data = {"token": token}
        if token_type is not None:
            data["token_type_hint"] = token_type
        url = self._requestor.reddit_url + const.REVOKE_TOKEN_PATH
        await self._post(url, success_status=codes["no_content"], **data)


class TrustedAuthenticator(BaseAuthenticator):
    """Store OAuth2 authentication credentials for web, or script type apps."""

    RESPONSE_TYPE = "code"

    def __init__(self, requestor, client_id, client_secret, redirect_uri=None):
        """Represent a single authentication to Reddit's API.

        :param requestor: An instance of :class:`Requestor`.
        :param client_id: The OAuth2 client ID to use with the session.
        :param client_secret: The OAuth2 client secret to use with the session.
        :param redirect_uri: (optional) The redirect URI exactly as specified
            in your OAuth application settings on Reddit. This parameter is
            required if you want to use the ``authorize_url`` method, or the
            ``authorize`` method of the ``Authorizer`` class.

        """
        super(TrustedAuthenticator, self).__init__(requestor, client_id, redirect_uri)
        self.client_secret = client_secret

    def _auth(self):
        return aiohttp.helpers.BasicAuth(self.client_id, self.client_secret)


class UntrustedAuthenticator(BaseAuthenticator):
    """Store OAuth2 authentication credentials for installed applications."""

    def _auth(self):
        return aiohttp.helpers.BasicAuth(self.client_id, "")


class BaseAuthorizer(object):
    """Superclass for OAuth2 authorization tokens and scopes."""

    def __init__(self, authenticator):
        """Represent a single authorization to Reddit's API.

        :param authenticator: An instance of :class:`BaseAuthenticator`.

        """
        self._authenticator = authenticator
        self._clear_access_token()
        self._validate_authenticator()

    def _clear_access_token(self):
        self._expiration_timestamp = None
        self.access_token = None
        self.scopes = None

    async def _request_token(self, **data):
        url = self._authenticator._requestor.reddit_url + const.ACCESS_TOKEN_PATH
        pre_request_time = time.time()
        response = await self._authenticator._post(url, **data)
        payload = await response.json()
        if "error" in payload:  # Why are these OKAY responses?
            raise OAuthException(
                response, payload["error"], payload.get("error_description")
            )

        self._expiration_timestamp = pre_request_time - 10 + payload["expires_in"]
        self.access_token = payload["access_token"]
        if "refresh_token" in payload:
            self.refresh_token = payload["refresh_token"]
        self.scopes = set(payload["scope"].split(" "))

    def _validate_authenticator(self):
        if not isinstance(self._authenticator, self.AUTHENTICATOR_CLASS):
            raise InvalidInvocation(
                f"Must use a authenticator of type {self.AUTHENTICATOR_CLASS.__name__}."
            )

    def is_valid(self):
        """Return whether or not the Authorizer is ready to authorize requests.

        A ``True`` return value does not guarantee that the access_token is
        actually valid on the server side.

        """
        return (
            self.access_token is not None and time.time() < self._expiration_timestamp
        )

    async def revoke(self):
        """Revoke the current Authorization."""
        if self.access_token is None:
            raise InvalidInvocation("no token available to revoke")

        await self._authenticator.revoke_token(self.access_token, "access_token")
        self._clear_access_token()


class Authorizer(BaseAuthorizer):
    """Manages OAuth2 authorization tokens and scopes."""

    AUTHENTICATOR_CLASS = BaseAuthenticator

    def __init__(
        self,
        authenticator,
        *,
        post_refresh_callback=None,
        pre_refresh_callback=None,
        refresh_token=None,
    ):
        """Represent a single authorization to Reddit's API.

        :param authenticator: An instance of a subclass of
            :class:`BaseAuthenticator`.
        :param post_refresh_callback: (Optional) When a single-argument synchronous or
            asynchronous function is passed, the function will be called prior to
            refreshing the access and refresh tokens. The argument to the callback is
            the :class:`Authorizer` instance. This callback can be used to inspect and
            modify the attributes of the :class:`Authorizer`.
        :param pre_refresh_callback: (Optional) When a single-argument function
            synchronous or asynchronous is passed, the function will be called after
            refreshing the access and refresh tokens. The argument to the callback is
            the :class:`Authorizer` instance. This callback can be used to inspect and
            modify the attributes of the :class:`Authorizer`.
        :param refresh_token: (Optional) Enables the ability to refresh the
            authorization.

        """
        super(Authorizer, self).__init__(authenticator)
        self._post_refresh_callback = post_refresh_callback
        self._pre_refresh_callback = pre_refresh_callback
        self.refresh_token = refresh_token

    async def authorize(self, code):
        """Obtain and set authorization tokens based on ``code``.

        :param code: The code obtained by an out-of-band authorization request
            to Reddit.

        """
        if self._authenticator.redirect_uri is None:
            raise InvalidInvocation("redirect URI not provided")
        await self._request_token(
            code=code,
            grant_type="authorization_code",
            redirect_uri=self._authenticator.redirect_uri,
        )

    async def refresh(self):
        """Obtain a new access token from the refresh_token."""
        if self._pre_refresh_callback:
            result = self._pre_refresh_callback(self)
            if inspect.isawaitable(result):
                await self._pre_refresh_callback(self)
        if self.refresh_token is None:
            raise InvalidInvocation("refresh token not provided")
        await self._request_token(
            grant_type="refresh_token", refresh_token=self.refresh_token
        )
        if self._post_refresh_callback:
            result = self._post_refresh_callback(self)
            if inspect.isawaitable(result):
                await self._post_refresh_callback(self)

    async def revoke(self, only_access=False):
        """Revoke the current Authorization.

        :param only_access: (Optional) When explicitly set to True, do not
            evict the refresh token if one is set.

        Revoking a refresh token will in-turn revoke all access tokens
        associated with that authorization.

        """
        if only_access or self.refresh_token is None:
            await super(Authorizer, self).revoke()
        else:
            await self._authenticator.revoke_token(self.refresh_token, "refresh_token")
            self._clear_access_token()
            self.refresh_token = None


class DeviceIDAuthorizer(BaseAuthorizer):
    """Manages app-only OAuth2 for 'installed' applications.

    While the '*' scope will be available, some endpoints simply will not work
    due to the lack of an associated Reddit account.

    """

    AUTHENTICATOR_CLASS = UntrustedAuthenticator

    def __init__(self, authenticator, device_id="DO_NOT_TRACK_THIS_DEVICE"):
        """Represent an app-only OAuth2 authorization for 'installed' apps.

        :param authenticator: An instance of :class:`UntrustedAuthenticator`.
        :param device_id: (optional) A unique ID (20-30 character ASCII string)
            (default DO_NOT_TRACK_THIS_DEVICE). For more information about this
            parameter, see:
            https://github.com/reddit/reddit/wiki/OAuth2#application-only-oauth
        """
        super(DeviceIDAuthorizer, self).__init__(authenticator)
        self._device_id = device_id

    async def refresh(self):
        """Obtain a new access token."""
        grant_type = "https://oauth.reddit.com/grants/installed_client"
        await self._request_token(grant_type=grant_type, device_id=self._device_id)


class ImplicitAuthorizer(BaseAuthorizer):
    """Manages implicit installed-app type authorizations."""

    AUTHENTICATOR_CLASS = UntrustedAuthenticator

    def __init__(self, authenticator, access_token, expires_in, scope):
        """Represent a single implicit authorization to Reddit's API.

        :param authenticator: An instance of :class:`UntrustedAuthenticator`.
        :param access_token: The access_token obtained from Reddit via callback
            to the authenticator's redirect_uri.
        :param expires_in: The number of seconds the ``access_token`` is valid
            for. The origin of this value was returned from Reddit via callback
            to the authenticator's redirect uri. Note, you may need to subtract
            an offset before passing in this number to account for a delay
            between when Reddit prepared the response, and when you make this
            function call.
        :param scope: A space-delimited string of Reddit OAuth2 scope names as
            returned from Reddit in the callback to the authenticator's
            redirect uri.

        """
        super(ImplicitAuthorizer, self).__init__(authenticator)
        self._expiration_timestamp = time.time() + expires_in
        self.access_token = access_token
        self.scopes = set(scope.split(" "))


class ReadOnlyAuthorizer(Authorizer):
    """Manages authorizations that are not associated with a Reddit account.

    While the '*' scope will be available, some endpoints simply will not work
    due to the lack of an associated Reddit account.

    """

    AUTHENTICATOR_CLASS = TrustedAuthenticator

    async def refresh(self):
        """Obtain a new ReadOnly access token."""
        await self._request_token(grant_type="client_credentials")


class ScriptAuthorizer(Authorizer):
    """Manages personal-use script type authorizations.

    Only users who are listed as developers for the application will be granted access
    tokens.

    """

    AUTHENTICATOR_CLASS = TrustedAuthenticator

    def __init__(self, authenticator, username, password):
        """Represent a single personal-use authorization to Reddit's API.

        :param authenticator: An instance of :class:`TrustedAuthenticator`.
        :param username: The Reddit username of one of the application's developers.
        :param password: The password associated with ``username``.

        """
        super(ScriptAuthorizer, self).__init__(authenticator)
        self._username = username
        self._password = password

    async def refresh(self):
        """Obtain a new personal-use script type access token."""
        await self._request_token(
            grant_type="password",
            username=self._username,
            password=self._password,
        )



"""Test for asyncprawcore.requestor.Requestor class."""
import asyncprawcore
import asynctest
import asyncio
from mock import patch, Mock
from asyncprawcore import RequestException


class RequestorTest(asynctest.TestCase):
    async def tearDown(self) -> None:
        if hasattr(self, "requestor"):
            if isinstance(self.requestor, asyncprawcore.requestor.Requestor):
                if not isinstance(self.requestor._http, Mock):
                    await self.requestor.close()

    async def test_initialize(self):
        self.requestor = asyncprawcore.Requestor("asyncprawcore:test (by /u/bboe)")
        self.assertEqual(
            f"asyncprawcore:test (by /u/bboe) asyncprawcore/{asyncprawcore.__version__}",
            self.requestor._http._default_headers["User-Agent"],
        )

    def test_initialize__failures(self):
        for agent in [None, "shorty"]:
            with self.assertRaises(asyncprawcore.InvalidInvocation):
                self.requestor = asyncprawcore.Requestor(agent)

    @patch("aiohttp.ClientSession")
    async def test_request__wrap_request_exceptions(self, mock_session):
        exception = Exception("asyncprawcore wrap_request_exceptions")
        session_instance = mock_session.return_value
        session_instance.request.side_effect = exception
        self.requestor = asyncprawcore.Requestor("asyncprawcore:test (by /u/bboe)")
        with self.assertRaises(asyncprawcore.RequestException) as context_manager:
            await self.requestor.request("get", "http://a.b", data="bar")
        self.assertIsInstance(context_manager.exception, RequestException)
        self.assertIs(exception, context_manager.exception.original_exception)
        self.assertEqual(("get", "http://a.b"), context_manager.exception.request_args)
        self.assertEqual({"data": "bar"}, context_manager.exception.request_kwargs)

    async def test_request__use_custom_session(self):
        override = "REQUEST OVERRIDDEN"
        custom_header = "CUSTOM SESSION HEADER"
        headers = {"session_header": custom_header}
        return_of_request = asyncio.Future()
        return_of_request.set_result(override)
        attrs = {
            "request.return_value": return_of_request,
            "_default_headers": headers,
        }
        session = Mock(**attrs)

        self.requestor = asyncprawcore.Requestor(
            "asyncprawcore:test (by /u/bboe)", session=session
        )

        self.assertEqual(
            f"asyncprawcore:test (by /u/bboe) asyncprawcore/{asyncprawcore.__version__}",
            self.requestor._http._default_headers["User-Agent"],
        )
        self.assertEqual(
            self.requestor._http._default_headers["session_header"],
            custom_header,
        )
        self.assertEqual(await self.requestor.request("https://reddit.com"), override)
